@@ -58,7 +58,12 @@ static DBusHandlerResult gs_listener_message_handler    (DBusConnection  *connec
 #define GS_LISTENER_PATH      "/org/gnome/ScreenSaver"
 #define GS_LISTENER_INTERFACE "org.gnome.ScreenSaver"
 
-#define DBUS_INTERFACE_HAL "org.freedesktop.Hal.Device"
+#define HAL_DEVICE_INTERFACE "org.freedesktop.Hal.Device"
+
+#define CK_NAME              "org.freedesktop.ConsoleKit"
+#define CK_MANAGER_PATH      "/org/freedesktop/ConsoleKit/Manager"
+#define CK_MANAGER_INTERFACE "org.freedesktop.ConsoleKit.Manager"
+#define CK_SESSION_INTERFACE "org.freedesktop.ConsoleKit.Session"
 
 #define TYPE_MISMATCH_ERROR GS_LISTENER_INTERFACE ".TypeMismatch"
 
@@ -77,6 +82,9 @@ struct GSListenerPrivate
         GHashTable     *throttlers;
         time_t          active_start;
         time_t          session_idle_start;
+        char           *session_id;
+
+        guint32         ck_throttle_cookie;
 };
 
 typedef struct
@@ -86,6 +94,7 @@ typedef struct
         char    *reason;
         char    *connection;
         guint32  cookie;
+        GTimeVal since;
 } GSListenerRefEntry;
 
 enum {
@@ -174,6 +183,30 @@ send_dbus_message (DBusConnection *connection,
         return sent;
 }
 
+static void
+send_dbus_boolean_signal (GSListener *listener,
+                          const char *name,
+                          gboolean    value)
+{
+        DBusMessage    *message;
+        DBusMessageIter iter;
+
+        g_return_if_fail (listener != NULL);
+
+        message = dbus_message_new_signal (GS_LISTENER_PATH,
+                                           GS_LISTENER_SERVICE,
+                                           name);
+
+        dbus_message_iter_init_append (message, &iter);
+        dbus_message_iter_append_basic (&iter, DBUS_TYPE_BOOLEAN, &value);
+
+        if (! send_dbus_message (listener->priv->connection, message)) {
+                gs_debug ("Could not send %s signal", name);
+        }
+
+        dbus_message_unref (message);
+}
+
 static void
 send_dbus_void_signal (GSListener *listener,
                        const char *name)
@@ -193,6 +226,15 @@ send_dbus_void_signal (GSListener *listener,
         dbus_message_unref (message);
 }
 
+void
+gs_listener_emit_power_notice (GSListener *listener,
+                               gboolean    in_effect)
+{
+        g_return_if_fail (listener != NULL);
+
+        send_dbus_boolean_signal (listener, "SessionPowerManagementIdleChanged", in_effect);
+}
+
 void
 gs_listener_emit_auth_request_begin (GSListener *listener)
 {
@@ -212,52 +254,74 @@ gs_listener_emit_auth_request_end (GSListener *listener)
 static void
 gs_listener_send_signal_active_changed (GSListener *listener)
 {
-        DBusMessage    *message;
-        DBusMessageIter iter;
-        dbus_bool_t     active;
-
         g_return_if_fail (listener != NULL);
 
         gs_debug ("Sending the ActiveChanged(%s) signal on the session bus",
                   listener->priv->active ? "TRUE" : "FALSE");
 
-        message = dbus_message_new_signal (GS_LISTENER_PATH,
-                                           GS_LISTENER_SERVICE,
-                                           "ActiveChanged");
-
-        active = listener->priv->active;
-        dbus_message_iter_init_append (message, &iter);
-        dbus_message_iter_append_basic (&iter, DBUS_TYPE_BOOLEAN, &active);
-
-        if (! send_dbus_message (listener->priv->connection, message)) {
-                gs_debug ("Could not send ActiveChanged signal");
-        }
-
-        dbus_message_unref (message);
+        send_dbus_boolean_signal (listener, "ActiveChanged", listener->priv->active);
 }
 
 static void
 gs_listener_send_signal_session_idle_changed (GSListener *listener)
+{
+        g_return_if_fail (listener != NULL);
+
+        send_dbus_boolean_signal (listener, "SessionIdleChanged", listener->priv->session_idle);
+}
+
+static void
+gs_listener_update_console_kit_idle (GSListener *listener)
 {
         DBusMessage    *message;
+        DBusMessage    *reply;
+        DBusError       error;
         DBusMessageIter iter;
-        dbus_bool_t     idle;
+        gboolean        idle;
 
         g_return_if_fail (listener != NULL);
 
-        message = dbus_message_new_signal (GS_LISTENER_PATH,
-                                           GS_LISTENER_SERVICE,
-                                           "SessionIdleChanged");
+        if (listener->priv->session_id == NULL) {
+                /* we don't have ConsoleKit running booo! */
+                return;
+        }
+
+        if (listener->priv->system_connection == NULL) {
+                gs_debug ("No connection to the system bus");
+                return;
+        }
 
         idle = listener->priv->session_idle;
+
+	gs_debug ("Updating ConsoleKit idle status: %d", idle);
+	message = dbus_message_new_method_call (CK_NAME,
+						listener->priv->session_id,
+						CK_SESSION_INTERFACE,
+						"SetIdleHint");
+	if (message == NULL) {
+		gs_debug ("Couldn't allocate the D-Bus message");
+		return;
+	}
+
         dbus_message_iter_init_append (message, &iter);
         dbus_message_iter_append_basic (&iter, DBUS_TYPE_BOOLEAN, &idle);
 
-        if (! send_dbus_message (listener->priv->connection, message)) {
-                gs_debug ("Could not send SessionIdleChanged signal");
-        }
-
+        /* FIXME: use async? */
+        dbus_error_init (&error);
+        reply = dbus_connection_send_with_reply_and_block (listener->priv->system_connection,
+                                                           message,
+                                                           -1, &error);
         dbus_message_unref (message);
+
+	if (reply != NULL) {
+                dbus_message_unref (reply);
+	}
+
+        if (dbus_error_is_set (&error)) {
+                gs_debug ("%s raised:\n %s\n\n", error.name, error.message);
+                dbus_error_free (&error);
+                return;
+        }
 }
 
 static const char *
@@ -412,6 +476,8 @@ listener_set_session_idle_internal (GSListener *listener,
 
         gs_listener_send_signal_session_idle_changed (listener);
 
+        gs_listener_update_console_kit_idle (listener);
+
         return TRUE;
 }
 
@@ -693,6 +759,124 @@ listener_remove_ref_entry (GSListener *listener,
         return removed;
 }
 
+#if GLIB_CHECK_VERSION(2,12,0)
+#define _g_time_val_to_iso8601(t) g_time_val_to_iso8601(t)
+#else
+/* copied from GLib */
+static gchar *
+_g_time_val_to_iso8601 (GTimeVal *time_)
+{
+  gchar *retval;
+
+  g_return_val_if_fail (time_->tv_usec >= 0 && time_->tv_usec < G_USEC_PER_SEC, NULL);
+
+#define ISO_8601_LEN 	21
+#define ISO_8601_FORMAT "%Y-%m-%dT%H:%M:%SZ"
+  retval = g_new0 (gchar, ISO_8601_LEN + 1);
+
+  strftime (retval, ISO_8601_LEN,
+	    ISO_8601_FORMAT,
+	    gmtime (&(time_->tv_sec)));
+
+  return retval;
+}
+#endif
+
+static void
+accumulate_ref_entry (gpointer            key,
+                      GSListenerRefEntry *entry,
+                      DBusMessageIter    *iter)
+{
+        char *description;
+        char *time;
+
+        time = _g_time_val_to_iso8601 (&entry->since);
+
+        description = g_strdup_printf ("Application=\"%s\"; Since=\"%s\"; Reason=\"%s\";",
+                                       entry->application,
+                                       time,
+                                       entry->reason);
+
+	dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &description);
+
+        g_free (description);
+        g_free (time);
+}
+
+static DBusHandlerResult
+listener_dbus_get_ref_entries (GSListener     *listener,
+                               int             entry_type,
+                               DBusConnection *connection,
+                               DBusMessage    *message)
+{
+        DBusMessage        *reply;
+        GHashTable         *hash;
+        DBusMessageIter     iter;
+	DBusMessageIter     iter_array;
+
+        hash = get_hash_for_entry_type (listener, entry_type);
+
+        reply = dbus_message_new_method_return (message);
+        if (reply == NULL) {
+                g_error ("No memory");
+        }
+
+        dbus_message_iter_init_append (reply, &iter);
+	dbus_message_iter_open_container (&iter,
+					  DBUS_TYPE_ARRAY,
+					  DBUS_TYPE_STRING_AS_STRING,
+					  &iter_array);
+
+        if (hash != NULL) {
+                g_hash_table_foreach (hash,
+                                      (GHFunc)accumulate_ref_entry,
+                                      &iter_array);
+        }
+
+	dbus_message_iter_close_container (&iter, &iter_array);
+
+        if (! dbus_connection_send (connection, reply, NULL)) {
+                g_error ("No memory");
+        }
+
+        dbus_message_unref (reply);
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+}
+
+static void
+listener_add_ck_ref_entry (GSListener     *listener,
+                           int             entry_type,
+                           DBusConnection *connection,
+                           DBusMessage    *message,
+                           guint32        *cookiep)
+{
+        GSListenerRefEntry *entry;
+
+        entry = g_new0 (GSListenerRefEntry, 1);
+        entry->entry_type = entry_type;
+        entry->connection = g_strdup (dbus_message_get_sender (message));
+        entry->cookie = listener_generate_unique_key (listener, entry_type);
+        entry->application = g_strdup ("ConsoleKit");
+        entry->reason = g_strdup ("Session is not active");
+        g_get_current_time (&entry->since);
+
+        /* takes ownership of entry */
+        listener_add_ref_entry (listener, entry_type, entry);
+
+        if (cookiep != NULL) {
+                *cookiep = entry->cookie;
+        }
+}
+
+static void
+listener_remove_ck_ref_entry (GSListener *listener,
+                              int         entry_type,
+                              guint32     cookie)
+{
+        listener_remove_ref_entry (listener, entry_type, cookie);
+}
+
 static DBusHandlerResult
 listener_dbus_add_ref_entry (GSListener     *listener,
                              int             entry_type,
@@ -733,6 +917,7 @@ listener_dbus_add_ref_entry (GSListener     *listener,
         entry->cookie = listener_generate_unique_key (listener, entry_type);
         entry->application = g_strdup (application);
         entry->reason = g_strdup (reason);
+        g_get_current_time (&entry->since);
 
         listener_add_ref_entry (listener, entry_type, entry);
 
@@ -1121,6 +1306,9 @@ do_introspect (DBusConnection *connection,
                                "    <method name=\"UnInhibit\">\n"
                                "      <arg name=\"cookie\" direction=\"in\" type=\"u\"/>\n"
                                "    </method>\n"
+                               "    <method name=\"GetInhibitors\">\n"
+                               "      <arg name=\"list\" direction=\"out\" type=\"as\"/>\n"
+                               "    </method>\n"
                                "    <method name=\"Throttle\">\n"
                                "      <arg name=\"application_name\" direction=\"in\" type=\"s\"/>\n"
                                "      <arg name=\"reason\" direction=\"in\" type=\"s\"/>\n"
@@ -1150,6 +1338,9 @@ do_introspect (DBusConnection *connection,
                                "    <signal name=\"SessionIdleChanged\">\n"
                                "      <arg name=\"new_value\" type=\"b\"/>\n"
                                "    </signal>\n"
+                               "    <signal name=\"SessionPowerManagementIdleChanged\">\n"
+                               "      <arg name=\"new_value\" type=\"b\"/>\n"
+                               "    </signal>\n"
                                "    <signal name=\"AuthenticationRequestBegin\">\n"
                                "    </signal>\n"
                                "    <signal name=\"AuthenticationRequestEnd\">\n"
@@ -1217,6 +1408,9 @@ listener_dbus_handle_session_message (DBusConnection *connection,
         if (dbus_message_is_method_call (message, GS_LISTENER_SERVICE, "UnInhibit")) {
                 return listener_dbus_remove_ref_entry (listener, REF_ENTRY_TYPE_INHIBIT, connection, message);
         }
+        if (dbus_message_is_method_call (message, GS_LISTENER_SERVICE, "GetInhibitors")) {
+                return listener_dbus_get_ref_entries (listener, REF_ENTRY_TYPE_INHIBIT, connection, message);
+        }
         if (dbus_message_is_method_call (message, GS_LISTENER_SERVICE, "Throttle")) {
                 return listener_dbus_add_ref_entry (listener, REF_ENTRY_TYPE_THROTTLE, connection, message);
         }
@@ -1249,6 +1443,25 @@ listener_dbus_handle_session_message (DBusConnection *connection,
         return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
+static gboolean
+_listener_message_path_is_our_session (GSListener  *listener,
+                                       DBusMessage *message)
+{
+        const char *ssid;
+        gboolean    ours;
+
+        ours = FALSE;
+
+        ssid = dbus_message_get_path (message);
+        if (ssid != NULL
+            && listener->priv->session_id != NULL
+            && strcmp (ssid, listener->priv->session_id) == 0) {
+                ours = TRUE;
+        }
+
+        return ours;
+}
+
 static DBusHandlerResult
 listener_dbus_handle_system_message (DBusConnection *connection,
                                      DBusMessage    *message,
@@ -1260,7 +1473,15 @@ listener_dbus_handle_system_message (DBusConnection *connection,
         g_return_val_if_fail (connection != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
         g_return_val_if_fail (message != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
 
-        if (dbus_message_is_signal (message, DBUS_INTERFACE_HAL, "Condition")) {
+#if 1
+        gs_debug ("obj_path=%s interface=%s method=%s destination=%s",
+                  dbus_message_get_path (message),
+                  dbus_message_get_interface (message),
+                  dbus_message_get_member (message),
+                  dbus_message_get_destination (message));
+#endif
+
+        if (dbus_message_is_signal (message, HAL_DEVICE_INTERFACE, "Condition")) {
                 DBusError   error;
                 const char *event;
                 const char *keyname;
@@ -1276,10 +1497,78 @@ listener_dbus_handle_system_message (DBusConnection *connection,
                                 g_signal_emit (listener, signals [LOCK], 0);
                         }
                 }
+
                 if (dbus_error_is_set (&error)) {
                         dbus_error_free (&error);
                 }
 
+                return DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_signal (message, CK_SESSION_INTERFACE, "Unlock")) {
+                if (_listener_message_path_is_our_session (listener, message)) {
+                        gs_debug ("Console kit requested session unlock");
+                        gs_listener_set_active (listener, FALSE);
+                }
+
+                return DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_signal (message, CK_SESSION_INTERFACE, "Lock")) {
+                if (_listener_message_path_is_our_session (listener, message)) {
+                        gs_debug ("ConsoleKit requested session lock");
+                        g_signal_emit (listener, signals [LOCK], 0);
+                }
+
+                return DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_signal (message, CK_SESSION_INTERFACE, "ActiveChanged")) {
+		/* NB that `ActiveChanged' refers to the active
+		 * session in ConsoleKit terminology - ie which
+		 * session is currently displayed on the screen.
+		 * gnome-screensaver uses `active' to mean `is the
+		 * screensaver active' (ie, is the screen locked) but
+		 * that's not what we're referring to here.
+		 */
+
+                if (_listener_message_path_is_our_session (listener, message)) {
+			DBusError   error;
+			dbus_bool_t new_active;
+
+			dbus_error_init (&error);
+			if (dbus_message_get_args (message, &error,
+						   DBUS_TYPE_BOOLEAN, &new_active,
+						   DBUS_TYPE_INVALID)) {
+				gs_debug ("ConsoleKit notified ActiveChanged %d", new_active);
+
+                                /* when we aren't active add an implicit throttle from CK
+                                 * when we become active remove the throttle and poke the lock */
+				if (new_active) {
+                                        if (listener->priv->ck_throttle_cookie != 0) {
+                                                listener_remove_ck_ref_entry (listener,
+                                                                              REF_ENTRY_TYPE_THROTTLE,
+                                                                              listener->priv->ck_throttle_cookie);
+                                                listener->priv->ck_throttle_cookie = 0;
+                                        }
+
+					g_signal_emit (listener, signals [SIMULATE_USER_ACTIVITY], 0);
+				} else {
+                                        if (listener->priv->ck_throttle_cookie != 0) {
+                                                g_warning ("ConsoleKit throttle already set");
+                                                listener_remove_ck_ref_entry (listener,
+                                                                              REF_ENTRY_TYPE_THROTTLE,
+                                                                              listener->priv->ck_throttle_cookie);
+                                                listener->priv->ck_throttle_cookie = 0;
+                                        }
+
+                                        listener_add_ck_ref_entry (listener,
+                                                                   REF_ENTRY_TYPE_THROTTLE,
+                                                                   connection,
+                                                                   message,
+                                                                   &listener->priv->ck_throttle_cookie);
+                                }
+			}
+
+			if (dbus_error_is_set (&error)) {
+				dbus_error_free (&error);
+			}
+                }
+
                 return DBUS_HANDLER_RESULT_HANDLED;
         }
 
@@ -1417,8 +1706,9 @@ listener_dbus_filter_function (DBusConnection *connection,
                                            DBUS_INTERFACE_DBUS,
                                            "NameOwnerChanged")) {
 
-                if (listener->priv->inhibitors != NULL)
+                if (listener->priv->inhibitors != NULL) {
                         listener_service_deleted (listener, message);
+                }
         } else {
                 return listener_dbus_handle_session_message (connection, message, user_data, FALSE);
         }
@@ -1687,13 +1977,80 @@ gs_listener_acquire (GSListener *listener,
 
         dbus_bus_add_match (listener->priv->system_connection,
                             "type='signal'"
-                            ",interface='"DBUS_INTERFACE_HAL"'"
+                            ",interface='"HAL_DEVICE_INTERFACE"'"
                             ",member='Condition'",
                             NULL);
+        dbus_bus_add_match (listener->priv->system_connection,
+                            "type='signal'"
+                            ",interface='"CK_SESSION_INTERFACE"'"
+                            ",member='Unlock'",
+                            NULL);
+        dbus_bus_add_match (listener->priv->system_connection,
+                            "type='signal'"
+                            ",interface='"CK_SESSION_INTERFACE"'"
+                            ",member='Lock'",
+                            NULL);
+        dbus_bus_add_match (listener->priv->system_connection,
+                            "type='signal'"
+                            ",interface='"CK_SESSION_INTERFACE"'"
+                            ",member='ActiveChanged'",
+                            NULL);
 
         return acquired;
 }
 
+static char *
+query_session_id (GSListener *listener)
+{
+        DBusMessage    *message;
+        DBusMessage    *reply;
+        DBusError       error;
+	DBusMessageIter reply_iter;
+        char           *ssid;
+
+        if (listener->priv->system_connection == NULL) {
+                gs_debug ("No connection to the system bus");
+                return NULL;
+        }
+
+        ssid = NULL;
+
+        dbus_error_init (&error);
+
+        message = dbus_message_new_method_call (CK_NAME, CK_MANAGER_PATH, CK_MANAGER_INTERFACE, "GetCurrentSession");
+        if (message == NULL) {
+                gs_debug ("Couldn't allocate the dbus message");
+                return NULL;
+        }
+
+        /* FIXME: use async? */
+        reply = dbus_connection_send_with_reply_and_block (listener->priv->system_connection,
+                                                           message,
+                                                           -1, &error);
+        dbus_message_unref (message);
+
+        if (dbus_error_is_set (&error)) {
+                gs_debug ("%s raised:\n %s\n\n", error.name, error.message);
+                dbus_error_free (&error);
+                return NULL;
+        }
+
+        dbus_message_iter_init (reply, &reply_iter);
+        dbus_message_iter_get_basic (&reply_iter, &ssid);
+
+        dbus_message_unref (reply);
+
+        return g_strdup (ssid);
+}
+
+static void
+init_session_id (GSListener *listener)
+{
+        g_free (listener->priv->session_id);
+        listener->priv->session_id = query_session_id (listener);
+        gs_debug ("Got session-id: %s", listener->priv->session_id);
+}
+
 static void
 gs_listener_init (GSListener *listener)
 {
@@ -1701,6 +2058,8 @@ gs_listener_init (GSListener *listener)
 
         gs_listener_dbus_init (listener);
 
+        init_session_id (listener);
+
         listener->priv->inhibitors = g_hash_table_new_full (g_int_hash,
                                                             g_int_equal,
                                                             NULL,
@@ -1731,6 +2090,8 @@ gs_listener_finalize (GObject *object)
                 g_hash_table_destroy (listener->priv->throttlers);
         }
 
+        g_free (listener->priv->session_id);
+
         G_OBJECT_CLASS (gs_listener_parent_class)->finalize (object);
 }
 
